/************************************************************************************		
 * Wzorzec:			fabryka abstrakcyjna (abstract factory)								   *
 *																												*
 * Zastosowanie: 	Wzorzec ten umoliwia atwe wprowadzanie zmian w zakresie klas    *
 *						uywanych w kodzie, co umoliwia pisanie kodu np: na kilka 		   *
 *						platform jednoczenie, czy atwiejsze zmienianie kodu wewntrz    *
 *						klasy bibliotecznej.															   *
 *																												*	
 *	Opis:				Fabryka abstrakcyjna, zwykle tworzona jako interfejs, w jzyku	   *
 *						c++ jest implementowana za pomoc klasy abstrakcyjnej. Klasa ta   *
 *						ma wygld typowej fabryki, jednak jako typy zwracane przez 		   *
 *						wirtualne metody fabrykujce ma zadeklarowane klasy abstrakcyjne 	*
 *						reprezentujce interfejsy obiektw zwracanych przez klasy 		   *
 *						pochodne. W kodzie programu implementuje si dla konkretnych 	   *
 *						zastosowa klasy pochodne od fabryki abstrakcyjnej wytwarzajce   *
 *						obiekty odpowiednie dla danego systemu czy pliku 					   *
 *						konfiguracyjnego. Dziki temu, e najczciej fabryka jest 			*
 *						tworzona w programie tylko raz, przy zmianie zachowania obiektw	*
 *						wystarczy zmodyfikowa t linijk w kodzie, w ktrej nastpuje 	*
 *						przypisanie konkretnej fabryki.												*
 *																												*
 *	Przykadowy schemat (schemat opisuje inny przypadek ni kod, dla urozmaicenia) :	*
 *	Prezentowany jest przypadek programowania dla kilku systemw operacyjnych			*
 *																												*
 *						Odczyt schematu:																	*
 *						[Element na pocztku strzaki]												*
 *						(d) dziedziczy po																	*
 *						(w) zawiera wskanik/wskaniki na											*
 *						(t) tworzy obiekty klasy														*
 *						[Element na ktry wskazuje strzaka]										*
 *					_______________				   ______________________						*
 *					| KlasaUsera  |-------------> |FabrykaAbstrakcyjna |						*
 *					|_____________|				(w)|____________________|						*
 *				   |_____________|					|____________________|						*
 *				   |_____________|					| A stworzObiekt()=0	|						*
 *							|								|____________________|						*
 *							|								  /|\(d)	/|\(d)  /|\(d)						*
 *							|				 ______________|		 |			|__________				*
 *							|		_______|__________ _________|________ ________|_________	*
 *							|		|FabrykaUnixowa  | |FabrykaDosowa	| |FabrykaWindows  |	*
 *	 -----------------/\----|________________| |________________| |________________|	*
 *	 |						|		|________________| |________________| |________________|	*
 *	 |						|		|A stworzObiekt()| |A stworzObiekt()| |A stworzObiekt()|	*
 *	 |					   |(w)	|________________| |________________| |________________|	*
 *  |		___________\|/_____									|						|				*
 *	 |		| A (klasa abstr)	|									|						|				*
 *	 |		|_________________|									|						|				*
 *	 |		|_________________|									|						|				*
 *	 |		| zareaguj()		|									|						|				*
 *	 |		|_________________|  								|						|				*
 *	 |		 /|\(d)	/|\(d)/|\(d)								|						|				*
*	 |(t)___|		 |		 |_________							|						|				*
 *	\|/__|______ ___|________ _____|______					|						|				*
 * | AUnix    | | ADos		| | AWindows |					|						|				*
 * |__________| |__________| |__________|	(t)			|						|				*
 *	|__________| |__________| |__________|<------------/\----------------|				*
 * |zareaguj()| |zareaguj()| |zareaguj()|					|										*
 *	|__________| |__________| |__________|					|										*
 *							/|\(t)									|										*
 *							 |_______________________________|										*
 *																												*
 ************************************************************************************/

/* Przykadowy kod. Przedstawia on rysowanie obiektw w zalenoci od
	szybkoci procesora */
#include <iostream>//Biblioteka potrzebna w tym przykadzie
using namespace std;// dla wygody, eby nie pisa wszdzie std

// Abstrakcyjny model obiektu graficznego
class ObiektGraficzny_Abstrakcyjny {
	protected:	
		float x;// pooenie na ekranie w osi x
		float y;// pooenie na ekranie w osi y
	public:
			virtual void rysujSie()=0;// metoda do rysowania obiekt
			// ponisze metody s wirtualne, gdy klasy pochodne nie musz koniecznie w ten sposb przechowywa
			// pooenia obiektu, jednak maj ju ciaa, gdy to bdzie najczstszy przypadek

			// metoda do ustalania pooenia obiektu
			virtual void ustawPozycje(float x, float y){
				this->x=x;
				this->y=y;			
			}
			// metody do pobierania pooenia obiektu
			virtual float podajPozX() {return x;}
			virtual float podajPozY() {return y;}
			ObiektGraficzny_Abstrakcyjny (float newX, float newY) : x(newX), y(newY){}
			virtual ~ObiektGraficzny_Abstrakcyjny() {}//wirtualny destruktor, bo s wirtualne funkcje
};

// Abstrakcyjna fabryka obiektw, suca za interfejs
class FabrykaObiektowGraficznych_Abstrakcyjna {
	public:
			// metoda do tworzenia instancji obiektw
			virtual ObiektGraficzny_Abstrakcyjny * utworzObiekt(float x, float y)=0;
			virtual ~FabrykaObiektowGraficznych_Abstrakcyjna(){}//wirtualny destruktor, bo sa wirtualne funkcje
}; 

// Konkretny model obiektu graficznego, dla wolniejszych komputerw
class ObiektGraficzny_Prosty : public ObiektGraficzny_Abstrakcyjny {
	public:
			// funkcja rysujSie jest mocno okrojona, gdy to tylko przykad zastosowania
			void rysujSie(){
				cout<<"malo wymagajacy model jest rysowany na pozycji: "<<x<<", "<<y<<endl;
			}
			ObiektGraficzny_Prosty (float x, float y):ObiektGraficzny_Abstrakcyjny(x,y){}
};

// Konkretny model obiektu graficznego, dla szybszych komputerw
class ObiektGraficzny_Trudny : public ObiektGraficzny_Abstrakcyjny {
	public:
			// funkcja rysujSie jest mocno okrojona, gdy to tylko przykad zastosowania
			void rysujSie(){
				cout<<"BARDZO WYMAGAJACY MODEL JEST RYSOWANY NA POZYCJI: "<<x<<", "<<y<<endl;
			}
			ObiektGraficzny_Trudny (float x, float y):ObiektGraficzny_Abstrakcyjny(x,y){}
};

//Konkretna fabryka obiektw graficznych, dla wolnych komputerw
class FabrykaObiektowGraficznych_Prostych : public FabrykaObiektowGraficznych_Abstrakcyjna {
	public:
		// skonkretyzowanie metody
		ObiektGraficzny_Abstrakcyjny * utworzObiekt(float x, float y) {
				return new ObiektGraficzny_Prosty(x,y);		
		}
};

//Konkretna fabryka obiektw graficznych, dla szybkich komputerw
class FabrykaObiektowGraficznych_Trudnych : public FabrykaObiektowGraficznych_Abstrakcyjna {
	public:
		// skonkretyzowanie metody
		ObiektGraficzny_Abstrakcyjny * utworzObiekt(float x, float y) {
				return new ObiektGraficzny_Trudny(x,y);		
		}
};

// Klasa korzystajca z fabryk
class KlasaUzytkowa {
	private:
		static const int ILE_OBIEKTOW=5;//ile obiektow graficznych naraz na ekranie maksymalnie
		// Wskaznik na fabryke obiektow, tu bedzie podany wskaznik do konkretnej fabryki 
	 	FabrykaObiektowGraficznych_Abstrakcyjna * fabryka;
		ObiektGraficzny_Abstrakcyjny * pulaObiektow[ILE_OBIEKTOW]; // Tablica obiektow do rysowania
	public:
		// dla zapeniania tablicy obiektw, zmienna nowa partia mwi czy zmienia obiekty,
		// czy te tworzy nowe od poczatku
		void pobierzNowe(bool nowaPartia){
			for (int licznik=0; licznik<5; ++licznik){
				ObiektGraficzny_Abstrakcyjny * temp=pulaObiektow[licznik];			
				if (temp!=NULL){// odtworzenie obiektow w nowych warunkach
						pulaObiektow[licznik]=fabryka->utworzObiekt(temp->podajPozX(), temp->podajPozY());	
						delete temp; 			
				}
				else if (nowaPartia){
					pulaObiektow[licznik]=fabryka->utworzObiekt(licznik, ILE_OBIEKTOW - licznik);	
				}				
			}		
		}
		// dla ustawienia rda obiektw
		void ustawFabryke(FabrykaObiektowGraficznych_Abstrakcyjna * nowaFabryka) {
				delete fabryka;// usunicie starej fabryki				
				fabryka=nowaFabryka;	// ustawienie nowej fabryki
				pobierzNowe(false);//eby od razu zmieni typ przechowywanych obiektw	
		} 
		// dla rysowania obiektw
		void rysujObiekty(){
			for (int licznik=0; licznik<ILE_OBIEKTOW; ++licznik){
				if (pulaObiektow[licznik]!=NULL){
						pulaObiektow[licznik]->rysujSie();				
				}			
			}		
		}
		// konstruktor
		KlasaUzytkowa (FabrykaObiektowGraficznych_Abstrakcyjna *nowaFabryka){
				fabryka=nowaFabryka;		
		}
		// destruktor
		~KlasaUzytkowa(){
			for (int licznik=0; licznik<5; ++licznik){
				delete pulaObiektow[licznik];// zwolnienie puli obiektw		
			}			
			delete fabryka;// zwolnienie fabryki
		}		
};

//Program testowy
int main(){
		cout<<"..:: Uruchamianie programu w trybie wysokiej jakosci grafiki ::.."<<endl;
		// Na pocztek ustawiany bardziej wymagajcy typ obiektu graficznego
		KlasaUzytkowa * testowa=new KlasaUzytkowa(new FabrykaObiektowGraficznych_Trudnych);
		testowa->pobierzNowe(true);// Ustawianie przykadowych obiektw
		cout<<"..:: Rysowanie przykladowych obiektow ::.."<<endl;
		testowa->rysujObiekty();
		//{...} Tutaj okazao si, e niestety maszyna jest za wolna
		cout<<"..:: Niestety maszyna jest za wolna, przelaczanie w tryb niskiej jakosci::.."<<endl;
		testowa->ustawFabryke(new FabrykaObiektowGraficznych_Prostych());
		cout<<"..:: Rysowanie przykladowych obiektow ::.."<<endl;
		testowa->rysujObiekty();
		cout<<"..:: Koniec programu ::.."<<endl;
}

